/*! ********************************************
 * SamplerPreset.as
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox.synth
{
/*
	- Parameters Notes -
	
	[toneName]
	  16 Chars (ASCII Code)
	  
	[smplNum]
	  SampleData Index Nuber [0 ~ 127]
	  
	[pan]
	  Panpot [ 0(L) ~ 64(center) ~ 127(R) ]
	
	[octv]
	  Tune (Octave)[0(-8 octave) ~ 8(+-0) ~ 15(+7 octave)]
	  
	[tune]
	  Tune [0(-1 octave) ~ 8192(+-0) ~ 16383(+1 octave)]
	  
	[tl]
	  Total Level [0 ~ 127]
	  0 -> 0dB
	  4 -> -3dB
	  8 -> -6dB
	  16 -> -12dB
	  127 -> -96dB
	  
	[ar]
	  Attack Rate [0 ~ 127]
	    0 -> AttackTime 256.0Sec / 1
	    8 -> AttackTime 256.0Sec / 2
	   16 -> AttackTime 256.0Sec / 4
	   24 -> AttackTime 256.0Sec / 8
	   32 -> AttackTime 256.0Sec / 16
	  128 -> AttackTime 256.0Sec / 65536
	  
	[dr]
	  Decay Rate [0 ~ 127]
	    0 -> DecayTime 256.0Sec / 1
	    8 -> DecayTime 256.0Sec / 2
	   16 -> DecayTime 256.0Sec / 4
	   24 -> DecayTime 256.0Sec / 8
	   32 -> DecayTime 256.0Sec / 16
	  128 -> DecayTime 256.0Sec / 65536
	  
	[sl]
	  Sustain Level [0 ~ 127]
	  0 -> 0dB
	  4 -> -3dB
	  8 -> -6dB
	  16 -> -12dB
	  127 -> -96dB
	  
	[rr]
	  Release Rate [0 ~ 127]
	    0 -> DecayTime 256.0Sec / 1
	    8 -> DecayTime 256.0Sec / 2
	   16 -> DecayTime 256.0Sec / 4
	   24 -> DecayTime 256.0Sec / 8
	   32 -> DecayTime 256.0Sec / 16
	  128 -> DecayTime 256.0Sec / 65536
	  
*/
public class SamplerPreset
{
	public var toneName	:String = "default preset";
	public var smplNum	:int = 0;    // 0 - 127
	public var pan		:int = 64;   // 0 - 127
	public var octv		:int = 8;    // 0 - 15     // center 8, octave = 1
	public var tune		:int = 8192; // 0 - 16383  // center 8192, octave = 8192
	public var tl		:int = 0;    // 0 - 127
	public var ar		:int = 127;  // 0 - 127
	public var dr		:int = 0;    // 0 - 127
	public var sl		:int = 127;  // 0 - 127
	public var rr		:int = 120;  // 0 - 127
	public function copy(preset:SamplerPreset):void
	{
		toneName = preset.toneName;
		smplNum = preset.smplNum;
		pan = preset.pan;
		octv = preset.octv;
		tune = preset.tune;
		tl = preset.tl;
		ar = preset.ar;
		dr = preset.dr;
		sl = preset.sl;
		rr = preset.rr;
	}
	public function reset():void
	{
		toneName = "default preset";
		smplNum = 0;
		pan = 64;
		octv = 8;
		tune = 8192;
		tl = 0;
		ar = 127;
		dr = 0;
		sl = 127;
		rr = 120;
	}
// Constructor
	public function SamplerPreset():void{}
}
}