/*! ********************************************
 * SynthesizerFM2.as
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox.synth
{
import flash.utils.ByteArray;
import sora.graybox.MidiEvent;
// 2op FM(PM) Synthesis
public final class SynthesizerFM2 extends MultiTimberSynthBase
{
// Factory -----
	override protected function _recreateVU(maxPoly:int):void
	{
		// Recreate Array
		if (_voiceUnits == null)_voiceUnits = new Vector.<VoiceUnitBase>(maxPoly, true);
		else 
		{
			_voiceUnits.fixed = false;
			_voiceUnits.length = maxPoly;
			_voiceUnits.fixed = true;
		}
		if (_voiceUnitExt == null)_voiceUnitExt = new Vector.<VoiceUnitFM2>(maxPoly, true);
		else 
		{
			_voiceUnitExt.fixed = false;
			_voiceUnitExt.length = maxPoly;
			_voiceUnitExt.fixed = true;
		}
		// Reset VoiceUnit
		var i:int, _vu:VoiceUnitFM2;
		for (i = 0; i < maxPoly; i++)
		{
			if (_voiceUnitExt[i] == null)
			{
				_voiceUnitExt[i] = new VoiceUnitFM2();
				_voiceUnitExt[i].setup(_synthData);
			}
			_voiceUnits[i] = _vu = _voiceUnitExt[i];
			_vu.killProcess();
			_vu.VOICE_PARTS = voiceParts;
			_vu.WAVE_TABLES = waveTables;
			_vu.SAMPLE_DATAS = sampleDatas;
			_vu.TONE_PRESETS = tonePresets;
			_vu.SAMPLER_PRESETS = samplerPresets;
			_vu.setMasterVolume(_masterVolume);
		}
		// set Property
		_maxPolyphony = maxPoly;
	}
	override protected function _resetKeyParams():void
	{
		// LFO
		var _lfoWaveSize:Number = SynthData.LFO_WAVE_TABLE_SIZE / 2.0,
			// Max Frame Rate - LFO Phase Step
			_lfoMaxFrameRate:Number = _sampleRate / SynthData.LFO_MAX_FREQ / _lfoWaveSize;
		if (_lfoMaxFrameRate < 1.0)_lfoMaxFrameRate = 1.0;  // Max 1frame per 1Phase
		var _lfoTimerCoef:int =  int(SynthData.LFO_TIMER_INIT_VALUE / 16383.0 / _lfoMaxFrameRate + 0.5);
		// EG
		var _rate0Frames:Number = SynthData.EG_REFERENCE_TIME * SynthData.LOG2VT_DYN_BITS * _sampleRate,
			_egPow2tIdx:int = SynthData.toPow2tIndexValue(SynthData.EG_MAX_CNT / _rate0Frames),
			_sndOffFrames:Number = SynthData.EG_SOUND_OFF_TIME / SynthData.LOG2VT_DYN_BITS * _sampleRate,
			_egSndOff:int = int(SynthData.EG_MAX_CNT / _sndOffFrames + 0.5);
		// PG PCM
		var _smpl:int = SynthData.toPow2tIndexValue(_sampleRate);
		var _pgSamplePow2tIdx:int = SynthData.toPow2tIndexValue(1 << PG_PHASE_FP) - _smpl;
		// PG WaveTable
		var _pgKeyPow2tIdx:int = SynthData.toPow2tIndexValue(WAVE_TABLE_SIZE << PG_PHASE_FP) - _smpl;
		_pgKeyPow2tIdx += SynthData.toPow2tIndexValue(_referencePitch * Math.pow(2.0, -69.0 / 12.0)); // note 0 Pitch
		var i:int, _vu:VoiceUnitFM2;
		for (i = 0; i < _maxPolyphony; i++)
		{
			_vu = _voiceUnitExt[i];
			_vu.killProcess();
			_vu.pgSamplePow2tIdx = _pgSamplePow2tIdx;
			_vu.pgKeyPow2tIdx = _pgKeyPow2tIdx;
			_vu.egPow2tIdx = _egPow2tIdx;
			_vu.egPhaseStepSndOff = _egSndOff;
			_vu.lfoTimerCoef = _lfoTimerCoef;
		}
	}
// Synthesizer - Data -----
	/*  
		Multiple
		MUL 0 - 1/2 pai       MUL 8 -  8 pai
		MUL 1 -   1 pai       MUL 9 -  9 pai
		MUL 2 -   2 pai       MUL10 - 10 pai
		MUL 3 -   3 pai       MUL11 - 11 pai
		MUL 4 -   4 pai       MUL12 - 12 pai
		MUL 5 -   5 pai       MUL13 - 13 pai
		MUL 6 -   6 pai       MUL14 - 14 pai
		MUL 7 -   7 pai       MUL15 - 15 pai
		* From Key Pitch
		
		Total Level
		         mod : car (pai)
		TL  0 -  32 : 1      TL 72 - 1 /  16 : 1 
		TL  8 -  16 : 1      TL 80 - 1 /  32 : 1
		TL 16 -   8 : 1      TL 88 - 1 /  64 : 1
		TL 24 -   4 : 1      TL 96 - 1 / 128 : 1
		TL 32 -   2 : 1      TL104 - 1 / 256 : 1
		TL 40 -   1 : 1      TL112 - 1 / 512 : 1
		TL 48 - 1/2 : 1      TL120 - 1 /1024 : 1
		TL 56 - 1/4 : 1      TL128 - 1 /2048 : 1  (table size)
		TL 64 - 1/8 : 1
		* Modulation Depth
		  LOG2VT output -> +-65536(131072)(Max) : (2048)WaveTableSize
	*/
	static public const WAVE_TABLE_SIZE_BITS :int = 11;                        // 16384 (DataEntry MSB LSB) / 8 (waves)
	static public const WAVE_TABLE_SIZE      :int = 1 << WAVE_TABLE_SIZE_BITS; // 16384 (DataEntry MSB LSB) / 8 (waves)
	static internal const PG_PHASE_FP :int = 16;  // 16bit fixed-point  (precision 1 / 65536)
	static internal const OVER_TONES  :int = 16;  // for Sinewave Synthesis.
	private var waveTables  :Vector.<Vector.<int>> = null;
	private var sampleDatas :Vector.<SampleData> = null;
	public var tonePresets :Vector.<TonePresetFM> = null;
	public var samplerPresets :Vector.<SamplerPreset> = null;
	private var _synthData:SynthData = null;
	public function getWaveForm(id:int, buffer:Vector.<Number>):void
	{
		if (buffer.length < WAVE_TABLE_SIZE) return;
		// ---
		var i:int, _table:Vector.<int> = waveTables[id];
		const _log2vt:Vector.<Number> = _synthData.getLOG2V_TABLE();
		for (i = 0; i < WAVE_TABLE_SIZE; i++) buffer[i] = _log2vt[_table[i]];
	}
	public function setWaveForm(id:int, buffer:Vector.<Number>):void
	{
		if (buffer.length < WAVE_TABLE_SIZE) return;
		const _LOGE:Number = -(SynthData.LOG2VT_DYN_RES * 1.442695040888963387),
			_BOADER:Number = 1.0 / (1 << SynthData.LOG2VT_DYN_BITS),
			_BOTTOM:int = SynthData.LOG2VT_SIZE;
		var i:int, v:Number, _table:Vector.<int> = waveTables[id];
		for (i = 0; i < WAVE_TABLE_SIZE; i++)
		{
			v = buffer[i];
			if      (v >  _BOADER) _table[i] =  int(Math.log( v) * _LOGE + 0.5) << 1;
			else if (v < -_BOADER) _table[i] = (int(Math.log(-v) * _LOGE + 0.5) << 1) + 1;
			else _table[i] = _BOTTOM;
		}
	}
	public function getSampleData(id:int):SampleData{ return sampleDatas[id]; }
	public function setSampleData(id:int, sampleData:SampleData, deepCopy:Boolean = true):void
	{
		allSoundStop();
		sampleDatas[id].copy(sampleData, deepCopy);
	}
	public function loadPCMSamples(id:int, wave:Vector.<Number>, ch:int):void
	{
		allSoundStop();
		sampleDatas[id].loadPCMSamples(wave, ch);
	}
	public function loadPCMData(id:int, dat:ByteArray, sampleFrames:int, ch:int, bitRate:int, sign:Boolean, bigEndian:Boolean):void
	{
		allSoundStop();
		sampleDatas[id].loadPCMData(dat, sampleFrames, ch, bitRate, sign, bigEndian);
	}
	public function clearSampleData(id:int):void
	{
		allSoundStop();
		sampleDatas[id].clear();
	}
	override public function factoryDataReset():void
	{
		if (_voiceUnitExt != null) allSoundStop();
		
		const _LOGE:Number = -(SynthData.LOG2VT_DYN_RES * 1.442695040888963387),
		      _BOADER:Number = 1.0 / (1 << SynthData.LOG2VT_DYN_BITS),
		      _BOTTOM:int = SynthData.LOG2VT_SIZE;
		var i:int;
	// Sample
		const _SIZE:int = WAVE_TABLE_SIZE, _HALF:int = _SIZE >> 1, _QTER:int = _SIZE >> 2,
			_PID2:Number = 3.141592653589793 * 2.0;
		var j:int, _nv:Number, _iv:int, _den:Number,
			_sl:Vector.<Number> = new Vector.<Number>(OVER_TONES, true),
			_sw:Vector.<Number> = new Vector.<Number>(OVER_TONES, true);
		
		if (waveTables == null) waveTables = new Vector.<Vector.<int>>(8, true);
		for (i = 0; i < 8; i++) if (waveTables[i] == null) waveTables[i] = new Vector.<int>(_SIZE, true);
		for (i = 0; i < _HALF; i++)
		{
		// sine [ nu ]
			_nv = Math.sin((i / _SIZE) * _PID2);
			if      (_nv >  _BOADER) _iv =  int(Math.log( _nv) * _LOGE + 0.5) << 1;
			else if (_nv < -_BOADER) _iv = (int(Math.log(-_nv) * _LOGE + 0.5) << 1) + 1;
			else _iv = _BOTTOM;
			waveTables[0][i] = _iv;
			waveTables[0][_HALF + i] = _iv + 1;
		// sine [ n- ]
			waveTables[1][i] = _iv;
			waveTables[1][_HALF + i] = _BOTTOM;
		// sine [ nn ]
			waveTables[2][i] = _iv; 
			waveTables[2][_HALF + i] = _iv;
			if	((i & 1) == 0){ 
		// sine [ nu-- ]
				waveTables[3][i >> 1] = _iv; 
				waveTables[3][_QTER + (i >> 1)] = _iv + 1;
		// sine [ nn-- ]
				waveTables[4][i >> 1] = _iv; 
				waveTables[4][_QTER + (i >> 1)] = _iv; }
			waveTables[3][_HALF + i] = _BOTTOM;
			waveTables[4][_HALF + i] = _BOTTOM;
		}
	// triangle ----
		for (i = 0, j = 1; i < OVER_TONES; i++, j+=2)_sw[i] = _SIZE / j;
		for (i = 0, j = 1; i < OVER_TONES; i++, j+=2)_sl[i] = (i & 1)? -1.0 / (j * j) : 1.0 / (j * j);
		_den = 0.0;
		for (j = 0; j < OVER_TONES; j++)_den += Math.sin(_QTER / _sw[j] * _PID2) * _sl[j];
		_den = 1.0 / _den;
		for (i = 0; i < _SIZE; i++)
		{
			_nv = 0.0;
			for (j = 0; j < OVER_TONES; j++)_nv += Math.sin((i / _sw[j]) * _PID2) * _sl[j];
			_nv *= _den;
			if      (_nv >  _BOADER) _iv =  int(Math.log( _nv) * _LOGE + 0.5) << 1;
			else if (_nv < -_BOADER) _iv = (int(Math.log(-_nv) * _LOGE + 0.5) << 1) + 1;
			else _iv = _BOTTOM;
			waveTables[5][i] = _iv;
		}
	// square ----
		//for (i = 0, j = 1; i < OVER_TONES; i++, j+=2)_sw[i] = _SIZE / j;
		for (i = 0, j = 1; i < OVER_TONES; i++, j+=2)_sl[i] = 1.0 / j;
		const _cQTR:Number = _QTER / OVER_TONES;
		_den = 0.0;
		for (j = 0; j < OVER_TONES; j++)_den += Math.sin(_cQTR / _sw[j] * _PID2) * _sl[j];
		_den = 1.0 / _den;
		for (i = 0; i < _SIZE; i++)
		{
			_nv = 0.0;
			for (j = 0; j < OVER_TONES; j++)_nv += Math.sin((i / _sw[j]) * _PID2) * _sl[j];
			_nv *= _den;
			if      (_nv >  _BOADER) _iv =  int(Math.log( _nv) * _LOGE + 0.5) << 1;
			else if (_nv < -_BOADER) _iv = (int(Math.log(-_nv) * _LOGE + 0.5) << 1) + 1;
			else _iv = _BOTTOM;
			waveTables[6][i] = _iv;
		}
	// saw ----
		for (i = 0, j = 1; i < OVER_TONES; i++, j++)_sw[i] = _SIZE / j;
		for (i = 0, j = 1; i < OVER_TONES; i++, j++)_sl[i] = 1.0 / j;
		var _c:Number = _HALF;
		for (i = 0, j = 1; i < OVER_TONES; i++, j++) _c /= 1.0 + 1.0 / j;
		_c = Math.round(_c + 0.5);
		_den = 0.0;
		for (j = 0; j < OVER_TONES; j++)_den += Math.sin(_c / _sw[j] * _PID2) * _sl[j];
		_den = 1.0 / _den;
		for (i = 0; i < _SIZE; i++)
		{
			_nv = 0.0;
			for (j = 0; j < OVER_TONES; j++)_nv += Math.sin((i / _sw[j]) * _PID2) * _sl[j];
			_nv *= _den;
			if      (_nv >  _BOADER) _iv =  int(Math.log( _nv) * _LOGE + 0.5) << 1;
			else if (_nv < -_BOADER) _iv = (int(Math.log(-_nv) * _LOGE + 0.5) << 1) + 1;
			else _iv = _BOTTOM;
			waveTables[7][i] = _iv;
		}
	// Preset
		if (tonePresets == null)tonePresets = new Vector.<TonePresetFM>(256, true);
		for (i = 0; i < 256; i++)
		{
			if (tonePresets[i] == null)tonePresets[i] = new TonePresetFM();
			tonePresets[i].reset();
		}
	// Sampler
		if (sampleDatas == null)sampleDatas = new Vector.<SampleData>(128, true);
		if (samplerPresets == null)samplerPresets = new Vector.<SamplerPreset>(128, true);
		for (i = 0; i < 128; i++)
		{
			if (sampleDatas[i] == null)sampleDatas[i] = new SampleData();
			else sampleDatas[i].clear();
			if (samplerPresets[i] == null)samplerPresets[i] = new SamplerPreset();
			else samplerPresets[i].reset();
		}
	}
// Synthesizer - main -----
	public const MODEL_ID :int = 0x21;
	//public    var  deviceID			:int 	= 0x10;
	private var _voiceUnitExt :Vector.<VoiceUnitFM2> = null;
// Constructor
	public function SynthesizerFM2(maxPoly:int = 32, refPitch:Number = 440.0, masVol:Number = 1.0, privateTableData:Boolean = true):void
	{
	// super ===
		super(maxPoly, refPitch, masVol);
	// SynthesizerFM2 ===
		_maxPMN = 3;
		// create
		_synthData = (privateTableData)? new SynthData() : SynthData.getInstance();
		factoryDataReset();
		_recreateVU(_maxPolyphony);
		_resetKeyParams();
	}
// MidiEvent ---
	override protected function _noteOn(ev:MidiEvent):void
	{
		var i:int, _vu:VoiceUnitFM2, _pt:VoicePart = voiceParts[ev.channel];
		for (i = 0; i < _maxPolyphony; i++)
		{
			_vu = _voiceUnitExt[i];
			if (_vu.state == 0)
			{
				_vu.startFrame = ev.offsetFrames;
				if (_pt.mode == 3)_vu.keyOnSampler(ev.channel, ev.value0, ev.value1);
				else _vu.keyOn(ev.channel, ev.value0, ev.value1);
				return;
			}
		}
		var _len:int = _nton + _maxPolyphony;
		for (i = _nton; i < _len; i++)
		{
			_vu = _voiceUnitExt[i & _ntonM];
			// (KeyOff || SoundOff) -->
			if (_vu.state & 0x60)
			{
				_nton = _nton + 1 & _ntonM;
				_vu.startFrame = ev.offsetFrames;
				if (_pt.mode == 3)_vu.keyOnSampler(ev.channel, ev.value0, ev.value1);
				else _vu.keyOn(ev.channel, ev.value0, ev.value1);
				return;
			}
		}
		for (i = _nton; i < _len; i++)
		{
			_vu = _voiceUnitExt[i & _ntonM];
			// KeyOn -->
			if (_vu.state & 0x10)
			{
				_nton = _nton + 1 & _ntonM;
				_vu.startFrame = ev.offsetFrames;
				if (_pt.mode == 3)_vu.keyOnSampler(ev.channel, ev.value0, ev.value1);
				else _vu.keyOn(ev.channel, ev.value0, ev.value1);
				return;
			}
		}
	}
	override protected function _readSysExData(ev:MidiEvent):void
	{
		var v:int, _data:ByteArray = ev.data;
		_data.position = 0;
		if (_data.readUnsignedShort() != 0x7D10) return;  // ManuFactureID, DeviceID
		if (_data.readByte()  		!=   0x21) return;  // modelID FM2
		switch (_data.readByte())	// Address 1
		{
		// Common
			case 0x00:
				switch (_data.readUnsignedShort())	// Address 2,3
				{
					case 0x0000 : setMasterVolume(_masterVolumeTable[_data.readByte()]); if (_data.bytesAvailable <= 1) break;
					case 0x0001 : setReferencePitch(440 + (_data.readByte() - 64) * 0.5);  break;
				}
				break;
		// Part
			case 0x02:
				v = _data.readByte() & 15;	// Address 2 [ CH ]
				switch (_data.readByte())	// Address 3
				{
					case 0x00 : setDefaultPartMode(v, _data.readByte()); if (_data.bytesAvailable <= 1) break;
					case 0x01 : setPartMode(v, _data.readByte()); break;
				}
				break;
		// Inst (Drum map)
			case 0x05: v = 128;
		// Inst (Program map)
			case 0x04:
				var _preset:TonePresetFM = tonePresets[v | _data.readByte()];	// Address 2 [ ID ]
				switch (_data.readByte())	// Address 3
				{
					case 0x00 : _preset.toneName = _data.readMultiByte(16, 'us-ascii'); if (_data.bytesAvailable <= 1) break;
					
					case 0x10 : _data.position++;  if (_data.bytesAvailable <= 1) break;
					case 0x11 : _data.position++;  if (_data.bytesAvailable <= 1) break;
					case 0x12 : _data.position++;  if (_data.bytesAvailable <= 1) break;
					case 0x13 : _data.position++;  if (_data.bytesAvailable <= 1) break;
					case 0x14 : _data.position++;  if (_data.bytesAvailable <= 1) break;
					case 0x15 : _data.position++;  if (_data.bytesAvailable <= 1) break;
					case 0x16 : _data.position++;  if (_data.bytesAvailable <= 1) break;
					case 0x17 : _data.position++;  if (_data.bytesAvailable <= 1) break;
					case 0x18 : _data.position++;  if (_data.bytesAvailable <= 1) break;
					case 0x19 : _preset.key     = _data.readByte();        if (_data.bytesAvailable <= 1) break;
					case 0x1A : _preset.pan     = _data.readByte();        if (_data.bytesAvailable <= 1) break;
					case 0x1B : _preset.lfoWave = _data.readByte() & 3;    if (_data.bytesAvailable <= 1) break;
					case 0x1C : _preset.lfoFreq = _preset.lfoFreq & 0x3f80 | _data.readByte();            if (_data.bytesAvailable <= 1) break;
					case 0x1D : _preset.lfoFreq = _preset.lfoFreq & 0x007f | _data.readByte() << 7;       if (_data.bytesAvailable <= 1) break;
					case 0x1E : _preset.fl      = _data.readByte();        if (_data.bytesAvailable <= 1) break;
					case 0x1F : _preset.strc    = _data.readByte() & 7;    if (_data.bytesAvailable <= 1) break;
					
					case 0x20 : _data.position++;							 if (_data.bytesAvailable <= 1) break;
					case 0x21 : _data.position++;							 if (_data.bytesAvailable <= 1) break;
					case 0x22 : _data.position++;							 if (_data.bytesAvailable <= 1) break;
					case 0x23 : _preset.tlVelCtrl_0 = _data.readByte() & 7;  if (_data.bytesAvailable <= 1) break;
					case 0x24 : _preset.mul_0       = _data.readByte() & 7;  if (_data.bytesAvailable <= 1) break;
					case 0x25 : _preset.tune_0      = _preset.tune_0 & 0x3f80 | _data.readByte();           if (_data.bytesAvailable <= 1) break;
					case 0x26 : _preset.tune_0      = _preset.tune_0 & 0x007f | _data.readByte() << 7;      if (_data.bytesAvailable <= 1) break;
					case 0x27 : _preset.lms_0       = _data.readByte() & 15; if (_data.bytesAvailable <= 1) break;
					case 0x28 : _preset.wave_0      = _data.readByte() & 7;  if (_data.bytesAvailable <= 1) break;
					case 0x29 : _preset.tl_0        = _data.readByte();      if (_data.bytesAvailable <= 1) break;
					case 0x2A : _preset.ar_0        = _data.readByte();      if (_data.bytesAvailable <= 1) break;
					case 0x2B : _preset.dr_0        = _data.readByte();      if (_data.bytesAvailable <= 1) break;
					case 0x2C : _preset.sl_0        = _data.readByte();      if (_data.bytesAvailable <= 1) break;
					case 0x2D : _preset.rr_0        = _data.readByte();      if (_data.bytesAvailable <= 1) break;
					case 0x2E : _preset.kf_0        = _data.readByte() & 3;  if (_data.bytesAvailable <= 1) break;
					case 0x2F : _preset.kfStart_0   = _data.readByte();      if (_data.bytesAvailable <= 1) break;
					
					case 0x30 : _data.position++;							 if (_data.bytesAvailable <= 1) break;
					case 0x31 : _data.position++;							 if (_data.bytesAvailable <= 1) break;
					case 0x32 : _data.position++;							 if (_data.bytesAvailable <= 1) break;
					case 0x33 : _preset.tlVelCtrl_1 = _data.readByte() & 7;  if (_data.bytesAvailable <= 1) break;
					case 0x34 : _preset.mul_1       = _data.readByte() & 7;  if (_data.bytesAvailable <= 1) break;
					case 0x35 : _preset.tune_1      = _preset.tune_1 & 0x3f80 | _data.readByte();           if (_data.bytesAvailable <= 1) break;
					case 0x36 : _preset.tune_1      = _preset.tune_1 & 0x007f | _data.readByte() << 7;      if (_data.bytesAvailable <= 1) break;
					case 0x37 : _preset.lms_1       = _data.readByte() & 15; if (_data.bytesAvailable <= 1) break;
					case 0x38 : _preset.wave_1      = _data.readByte() & 7;  if (_data.bytesAvailable <= 1) break;
					case 0x39 : _preset.tl_1        = _data.readByte();      if (_data.bytesAvailable <= 1) break;
					case 0x3A : _preset.ar_1        = _data.readByte();      if (_data.bytesAvailable <= 1) break;
					case 0x3B : _preset.dr_1        = _data.readByte();      if (_data.bytesAvailable <= 1) break;
					case 0x3C : _preset.sl_1        = _data.readByte();      if (_data.bytesAvailable <= 1) break;
					case 0x3D : _preset.rr_1        = _data.readByte();      if (_data.bytesAvailable <= 1) break;
					case 0x3E : _preset.kf_1        = _data.readByte() & 3;  if (_data.bytesAvailable <= 1) break;
					case 0x3F : _preset.kfStart_1   = _data.readByte();      break;
				}
				break;
		// Percussion
			case 0x06:
				var  _smplPrst:SamplerPreset = samplerPresets[_data.readByte()];	// Address 2 [ ID ]
				switch (_data.readByte())	// Address 3
				{
					case 0x00 : _smplPrst.toneName = _data.readMultiByte(16, 'us-ascii'); if (_data.bytesAvailable <= 1) break;
					case 0x10 : _smplPrst.smplNum = _data.readByte();   if (_data.bytesAvailable <= 1) break;
					case 0x11 : _smplPrst.pan = _data.readByte();       if (_data.bytesAvailable <= 1) break;
					case 0x12 : _smplPrst.octv = _data.readByte() & 15; if (_data.bytesAvailable <= 1) break;
					case 0x13 : _smplPrst.tune = _smplPrst.tune & 0x3f80 | _data.readByte();           if (_data.bytesAvailable <= 1) break;
					case 0x14 : _smplPrst.tune = _smplPrst.tune & 0x007f | _data.readByte() << 7;      if (_data.bytesAvailable <= 1) break;
					case 0x15 : _smplPrst.tl = _data.readByte();        if (_data.bytesAvailable <= 1) break;
					case 0x16 : _smplPrst.ar = _data.readByte();        if (_data.bytesAvailable <= 1) break;
					case 0x17 : _smplPrst.dr = _data.readByte();        if (_data.bytesAvailable <= 1) break;
					case 0x18 : _smplPrst.sl = _data.readByte();        if (_data.bytesAvailable <= 1) break;
					case 0x19 : _smplPrst.rr = _data.readByte();        break;
				}
				break;
		}
	}
}
}