/*! ********************************************
 * SynthesizerPCM.as
 * Copyright 2019 masahiro takamoto (MotChan)
 * Distributed under the MIT license. 
 * see -> sora/license.txt
 ******************************************** */
package sora.graybox.synth
{
import flash.utils.ByteArray;
import sora.graybox.MidiEvent;
// PCM WaveTable Synthesis
public final class SynthesizerPCM extends MultiTimberSynthBase
{
// Factory -----
	override protected function _recreateVU(maxPoly:int):void
	{
		// Recreate Array
		if (_voiceUnits == null)_voiceUnits = new Vector.<VoiceUnitBase>(maxPoly, true);
		else 
		{
			_voiceUnits.fixed = false;
			_voiceUnits.length = maxPoly;
			_voiceUnits.fixed = true;
		}
		if (_voiceUnitExt == null)_voiceUnitExt = new Vector.<VoiceUnitPCM>(maxPoly, true);
		else 
		{
			_voiceUnitExt.fixed = false;
			_voiceUnitExt.length = maxPoly;
			_voiceUnitExt.fixed = true;
		}
		// Reset VoiceUnit
		var i:int, _vu:VoiceUnitPCM;
		for (i = 0; i < maxPoly; i++)
		{
			if (_voiceUnitExt[i] == null)
			{
				_voiceUnitExt[i] = new VoiceUnitPCM();
				_voiceUnitExt[i].setup(_synthData);
			}
			_voiceUnits[i] = _vu = _voiceUnitExt[i];
			_vu.killProcess();
			_vu.VOICE_PARTS = voiceParts;
			_vu.SAMPLE_DATAS = sampleDatas;
			_vu.TONE_PRESETS = tonePresets;
			_vu.SAMPLER_PRESETS = samplerPresets;
			_vu.setMasterVolume(_masterVolume);
		}
		// set Property
		_maxPolyphony = maxPoly;
	}
	override protected function _resetKeyParams():void
	{
		// LFO
		var _lfoWaveSize:Number = SynthData.LFO_WAVE_TABLE_SIZE / 2.0,
			// Max Frame Rate - LFO Phase Step
			_lfoMaxFrameRate:Number = _sampleRate / SynthData.LFO_MAX_FREQ / _lfoWaveSize;
		if (_lfoMaxFrameRate < 1.0)_lfoMaxFrameRate = 1.0;  // Max 1frame per 1Phase
		var _lfoTimerCoef:int =  int(SynthData.LFO_TIMER_INIT_VALUE / 16383.0 / _lfoMaxFrameRate + 0.5);
		// EG
		var _rate0Frames:Number = SynthData.EG_REFERENCE_TIME * SynthData.LOG2VT_DYN_BITS * _sampleRate,
			_egPow2tIdx:int = SynthData.toPow2tIndexValue(SynthData.EG_MAX_CNT / _rate0Frames),
			_sndOffFrames:Number = SynthData.EG_SOUND_OFF_TIME / SynthData.LOG2VT_DYN_BITS * _sampleRate,
			_egSndOff:int = int(SynthData.EG_MAX_CNT / _sndOffFrames + 0.5);
		// PG PCM
		var _smpl:int = SynthData.toPow2tIndexValue(_sampleRate);
		var _pgSamplePow2tIdx:int = SynthData.toPow2tIndexValue(1 << PG_PHASE_FP) - _smpl;
		// PG WaveTable
		var _pgKeyPow2tIdx:int = _pgSamplePow2tIdx + SynthData.toPow2tIndexValue(_referencePitch * Math.pow(2.0, -69.0 / 12.0)); // note 0 Pitch
		var i:int, _vu:VoiceUnitPCM;
		for (i = 0; i < _maxPolyphony; i++)
		{
			_vu = _voiceUnitExt[i];
			_vu.killProcess();
			_vu.pgSamplePow2tIdx = _pgSamplePow2tIdx;
			_vu.pgKeyPow2tIdx = _pgKeyPow2tIdx;
			_vu.egPow2tIdx = _egPow2tIdx;
			_vu.egPhaseStepSndOff = _egSndOff;
			_vu.lfoTimerCoef = _lfoTimerCoef;
		}
	}
// Synthesizer - Data -----
	static internal const PG_PHASE_FP :int = 18;  // 18bit fixed-point  (precision 1 / 262144)
	private var sampleDatas :Vector.<SampleData> = null;
	public var tonePresets :Vector.<TonePresetPCM> = null;
	public var samplerPresets :Vector.<SamplerPreset> = null;
	private var _synthData:SynthData = null;
	public function getSampleData(id:int):SampleData{ return sampleDatas[id]; }
	public function setSampleData(id:int, sampleData:SampleData, deepCopy:Boolean = true):void
	{
		allSoundStop();
		sampleDatas[id].copy(sampleData, deepCopy);
	}
	public function loadPCMSamples(id:int, wave:Vector.<Number>, ch:int):void
	{
		allSoundStop();
		sampleDatas[id].loadPCMSamples(wave, ch);
	}
	public function loadPCMData(id:int, dat:ByteArray, sampleFrames:int, ch:int, bitRate:int, sign:Boolean, bigEndian:Boolean):void
	{
		allSoundStop();
		sampleDatas[id].loadPCMData(dat, sampleFrames, ch, bitRate, sign, bigEndian);
	}
	public function clearSampleData(id:int):void
	{
		allSoundStop();
		sampleDatas[id].clear();
	}
	override public function factoryDataReset():void
	{
		if (_voiceUnitExt != null) allSoundStop();
		var i:int;
	// Sample
		if (sampleDatas == null) sampleDatas = new Vector.<SampleData>(128, true);
		for (i = 0; i < 128; i++)
		{
			if (sampleDatas[i] == null) sampleDatas[i] = new SampleData();
			sampleDatas[i].clear();
		}
		if (tonePresets == null) tonePresets = new Vector.<TonePresetPCM>(128, true);
		if (samplerPresets == null) samplerPresets = new Vector.<SamplerPreset>(128, true);
		for (i = 0; i < 128; i++)
		{
			if (tonePresets[i] == null)tonePresets[i] = new TonePresetPCM();
			if (samplerPresets[i] == null)samplerPresets[i] = new SamplerPreset();
			tonePresets[i].reset();
			samplerPresets[i].reset();
		}
	}
// Synthesizer - main -----
	public const MODEL_ID :int = 0x23;
	//public    var  deviceID			:int 	= 0x10;
	private var _voiceUnitExt :Vector.<VoiceUnitPCM> = null;
// Constructor
	public function SynthesizerPCM(maxPoly:int = 32, refPitch:Number = 440.0, masVol:Number = 1.0, privateTableData:Boolean = true):void
	{
	// super ===
		super(maxPoly, refPitch, masVol);
	// SynthesizerPCM ===
		_maxPMN = 3;
		// create
		_synthData = (privateTableData)? new SynthData() : SynthData.getInstance();
		factoryDataReset();
		_recreateVU(_maxPolyphony);
		_resetKeyParams();
	}
// MidiEvent ---
	override protected function _noteOn(ev:MidiEvent):void
	{
		var i:int, _vu:VoiceUnitPCM, _pt:VoicePart = voiceParts[ev.channel];
		for (i = 0; i < _maxPolyphony; i++)
		{
			_vu = _voiceUnitExt[i];
			if (_vu.state == 0)
			{
				_vu.startFrame = ev.offsetFrames;
				if (_pt.mode == 3)_vu.keyOnSampler(ev.channel, ev.value0, ev.value1);
				else _vu.keyOn(ev.channel, ev.value0, ev.value1);
				return;
			}
		}
		var _len:int = _nton + _maxPolyphony;
		for (i = _nton; i < _len; i++)
		{
			_vu = _voiceUnitExt[i & _ntonM];
			// (KeyOff || SoundOff) -->
			if (_vu.state & 0x60)
			{
				_nton = _nton + 1 & _ntonM;
				_vu.startFrame = ev.offsetFrames;
				if (_pt.mode == 3)_vu.keyOnSampler(ev.channel, ev.value0, ev.value1);
				else _vu.keyOn(ev.channel, ev.value0, ev.value1);
				return;
			}
		}
		for (i = _nton; i < _len; i++)
		{
			_vu = _voiceUnitExt[i & _ntonM];
			// KeyOn -->
			if (_vu.state & 0x10)
			{
				_nton = _nton + 1 & _ntonM;
				_vu.startFrame = ev.offsetFrames;
				if (_pt.mode == 3)_vu.keyOnSampler(ev.channel, ev.value0, ev.value1);
				else _vu.keyOn(ev.channel, ev.value0, ev.value1);
				return;
			}
		}
	}
	override protected function _readSysExData(ev:MidiEvent):void
	{
		var v:int, _data:ByteArray = ev.data;
		_data.position = 0;
		if (_data.readUnsignedShort() != 0x7D10) return;  // ManuFactureID, DeviceID
		if (_data.readByte() 			!=   0x23) return;  // modelID PCM
		switch (_data.readByte())	// Address 1
		{
		// Common
			case 0x00:
				switch (_data.readUnsignedShort())	// Address 2,3
				{
					case 0x0000 : setMasterVolume(_masterVolumeTable[_data.readByte()]); if (_data.bytesAvailable <= 1) break;
					case 0x0001 : setReferencePitch(440 + (_data.readByte() - 64) * 0.5);  break;
				}
				break;
		// Part
			case 0x02:
				v = _data.readByte() & 15;	// Address 2 [ CH ]
				switch (_data.readByte())	// Address 3
				{
					case 0x00 : setDefaultPartMode(v, _data.readByte()); if (_data.bytesAvailable <= 1) break;
					case 0x01 : setPartMode(v, _data.readByte()); break;
				}
				break;
		// Inst
			case 0x04:
				var _preset:TonePresetPCM = tonePresets[_data.readByte()];	// Address 2 [ ID ]
				switch (_data.readByte())	// Address 3
				{
					case 0x00 : _preset.toneName = _data.readMultiByte(16, 'us-ascii');             if (_data.bytesAvailable <= 1) break;
					case 0x10 : _data.position++;                    if (_data.bytesAvailable <= 1) break;
					case 0x11 : _data.position++;                    if (_data.bytesAvailable <= 1) break;
					case 0x12 : _preset.lfoFreq = _preset.lfoFreq & 0x3f80 | _data.readByte();      if (_data.bytesAvailable <= 1) break;
					case 0x13 : _preset.lfoFreq = _preset.lfoFreq & 0x007f | _data.readByte() << 7; if (_data.bytesAvailable <= 1) break;
					case 0x14 : _preset.strc = _data.readByte() & 3; if (_data.bytesAvailable <= 1) break;
					case 0x15 : _preset.tune = _preset.tune & 0x3f80 | _data.readByte();            if (_data.bytesAvailable <= 1) break;
					case 0x16 : _preset.tune = _preset.tune & 0x007f | _data.readByte() << 7;       if (_data.bytesAvailable <= 1) break;
					case 0x17 : _preset.lms = _data.readByte() & 15; if (_data.bytesAvailable <= 1) break;
					case 0x18 : _preset.smplNum = _data.readByte();  if (_data.bytesAvailable <= 1) break;
					case 0x19 : _preset.tl = _data.readByte();       if (_data.bytesAvailable <= 1) break;
					case 0x1A : _preset.ar = _data.readByte();       if (_data.bytesAvailable <= 1) break;
					case 0x1B : _preset.dr = _data.readByte();       if (_data.bytesAvailable <= 1) break;
					case 0x1C : _preset.sl = _data.readByte();       if (_data.bytesAvailable <= 1) break;
					case 0x1D : _preset.rr = _data.readByte();       if (_data.bytesAvailable <= 1) break;
					case 0x1E : _preset.kf = _data.readByte() & 3;   if (_data.bytesAvailable <= 1) break;
					case 0x1F : _preset.kfStart = _data.readByte();  break;
				}
				break;
		// Percussion
			case 0x06:
				var  _smplPrst:SamplerPreset = samplerPresets[_data.readByte()];	// Address 2 [ ID ]
				switch (_data.readByte())	// Address 3
				{
					case 0x00 : _smplPrst.toneName = _data.readMultiByte(16, 'us-ascii');               if (_data.bytesAvailable <= 1) break;
					case 0x10 : _smplPrst.smplNum = _data.readByte();    if (_data.bytesAvailable <= 1) break;
					case 0x11 : _smplPrst.pan = _data.readByte();        if (_data.bytesAvailable <= 1) break;
					case 0x12 : _smplPrst.octv = _data.readByte() & 15;  if (_data.bytesAvailable <= 1) break;
					case 0x13 : _smplPrst.tune = _smplPrst.tune & 0x3f80 | _data.readByte();      if (_data.bytesAvailable <= 1) break;
					case 0x14 : _smplPrst.tune = _smplPrst.tune & 0x007f | _data.readByte() << 7; if (_data.bytesAvailable <= 1) break;
					case 0x15 : _smplPrst.tl = _data.readByte();         if (_data.bytesAvailable <= 1) break;
					case 0x16 : _smplPrst.ar = _data.readByte();         if (_data.bytesAvailable <= 1) break;
					case 0x17 : _smplPrst.dr = _data.readByte();         if (_data.bytesAvailable <= 1) break;
					case 0x18 : _smplPrst.sl = _data.readByte();         if (_data.bytesAvailable <= 1) break;
					case 0x19 : _smplPrst.rr = _data.readByte();         break;
				}
				break;
		}
	}
}
}